#include <iostream>
#include <limits>

#include "life.h"

void simulate_life_serial(LifeBoard &state, int steps) {
    LifeBoard next_state{state.width(), state.height()};
    for (int step = 0; step < steps; ++step) {
        /* We use the range [1, width - 1) here instead of
         * [0, width) because we fix the edges to be all 0s.
         */
        for (int y = 1; y < state.height() - 1; ++y) {
            for (int x = 1; x < state.width() - 1; ++x) {
                int live_in_window = 0;
                /* For each cell, examine a 3x3 "window" of cells around it,
                 * and count the number of live (true) cells in the window. */
                for (int y_offset = -1; y_offset <= 1; ++y_offset) {
                    for (int x_offset = -1; x_offset <= 1; ++x_offset) {
                        if (state.at(x + x_offset, y + y_offset)) {
                            live_in_window += 1;
                        }
                    }
                }
                if (state.at(x, y)) {
                    /* Live cells remain live only if they have 2 or 3 live neighbors */
                    next_state.at(x, y) = (live_in_window == 3 || live_in_window == 4);
                } else {
                    /* Dead cells become live only if they have exactly 3 live neighbors */
                    next_state.at(x, y) = (live_in_window == 3);
                }
            }
        }
        
        /* now that we computed next_state, make it the current state */
        swap(state, next_state);
    }
}

std::ostream& operator<<(std::ostream& out, LifeBoard const& state) {
    out << state.width() << " " << state.height() << "\n";
    for (int y = 0; y < state.height(); ++y) {
        for (int x = 0; x < state.width(); ++x) {
            if (state.at(x, y)) {
                out << '*';
            } else {
                out << '.';
            }
        }
        out << "\n";
    }
    return out;
}

std::istream& operator>>(std::istream& in, LifeBoard& state) {
    in >> state.width_ >> state.height_;
    in.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
    state.cells_.resize(state.width_ * state.height_);
    for (int y = 0; y < state.height(); ++y) {
        for (int x = 0; x < state.width(); ++x) {
            char c = in.get();
            state.at(x, y) = !(c == ' ' || c == '.');
        }
        in.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
    }
    return in;
}
